% -------------------------------------------------------------------------------------
% Comparison-Based Optimizers Need Comparison-Based Surrogates. PPSN 2010.
% ACM-ES and CMA_ES implementation is based on the source code of N. Hansen and O. Shir. 
% To be used under the terms of the GNU General Public License:
% http://www.gnu.org/copyleft/gpl.html
%
% Authors: Ilya Loshchilov, Marc Schoenauer and Michele Sebag, 2010. 
% E-mail: ilya.loshchilov@gmail.com , marc.schoenauer@inria.fr , michele.sebag@lri.fr
% Reference: http://www.lri.fr/~ilya/ , http://www.lri.fr/~marc/ , http://www.lri.fr/~sebag/
% -------------------------------------------------------------------------------------

function ACMES

N = 10;                 % dimension
stopfitness = 1e-10;    % stop if fitness < stopfitness (minimization)
useSVM = 1;             % 1 - ACM-ES; 0 - CMA-ES
printdelay = 50;        % every (nevals%printdelay) print 'bestfit'

% strfitnessfct = 'fspherenoise';    X_a = -5;   X_b = 5;       co_sigma = 0.5;      
% strfitnessfct = 'fschwefel';       X_a = -10;  X_b = 10;      co_sigma = 10;         
% strfitnessfct = 'fackley';         X_a = 1;    X_b = 30;      co_sigma = 14.5; 	
% strfitnessfct = 'fschwefel14';     X_a = -10;  X_b = 10;      co_sigma = 10;       
  strfitnessfct = 'frosenbrock';     X_a = -5;   X_b = 5;       co_sigma = 0.5;      
% strfitnessfct = 'felli';           X_a = 1;    X_b = 5;       co_sigma = 2;        
% strfitnessfct = 'frastrigin';      X_a = 1;    X_b = 5;       co_sigma = 2;       

MAX_EVAL = 3000 * N;  
if (useSVM == 1)    MAX_EVAL = 1500*N;  end;
 
tic;
[X , F, nevals] = cmaplus(strfitnessfct,N,X_a,X_b,co_sigma,MAX_EVAL,stopfitness, useSVM, printdelay );
toc

end

function [X , F, nevals] = cmaplus(strfitnessfct,N,X_a,X_b,sigma,MAX_EVAL,stopfitness, useSVM, printdelay);
close all;

%Strategy parameter settings: ACM-ES
p = 0.4;
sigma1 = p;
sigma2 = 2*p;

nTrain0 = 30;
if (strncmp(strfitnessfct,'frosenbrock',length(strfitnessfct)) == 1)
    nTrain0 = 70;
end;
if (strncmp(strfitnessfct,'felli',length(strfitnessfct)) == 1)
    nTrain0 = 70;
end;
p = 0.5;
nTrainMax = floor(nTrain0 *(N^p));
nTest = 500;
niter = floor(50000 * (N^p));%150000;
       
lambda = 4+floor(3*log(N));  % population size, offspring number
mu = floor(lambda/2); 

nfcalc = floor(lambda/3);
if (nfcalc < 3) nfcalc = 3; end;

if (strncmp(strfitnessfct,'frastrigin',length(strfitnessfct)) == 1)
    if (N == 2)     lambda = 50;    mu = 25;  end;
    if (N == 5)     lambda = 140;   mu = 70;  end;
    if (N == 10)    lambda = 500;    end;   
    nfcalc = floor(lambda/2);
if (nfcalc < 3) nfcalc = 3; end;
end;

%Strategy parameter setting: CMA-ES
weights = log(mu+1/2)-log(1:mu)';     % muXone array for weighted recombination
weights = weights/sum(weights);       % normalize recombination weights array

mueff=sum(weights)^2/sum(weights.^2); % variance-effectiveness of sum w_i x_i

cc = (4 + mueff/N) / (N+4 + 2*mueff/N); % time constant for cumulation for C
cs = (mueff+2) / (N+mueff+5);  % t-const for cumulation for sigma control
c1 = 2 / ((N+1.3)^2+mueff);    % learning rate for rank-one update of C
cmu = 2 * (mueff-2+1/mueff) / ((N+2)^2+mueff);  % and for rank-mu update
damps = 1 + 2*max(0, sqrt((mueff-1)/(N+1))-1) + cs; % damping for sigma 
                                                              % usually close to 1
% Initialize dynamic (internal) strategy parameters and constants
chiN=N^0.5*(1-1/(4*N)+1/(21*N^2));  % expectation of
ps = zeros(N,1);        % evolution paths for sigma

%Initial Point
Xmean = (X_b-X_a)*rand(N,1) + X_a; %decision parameters to be optimized.

% Initialize strategy parameters and constants
pc = zeros(N,1);                    % evolution paths for C
B = eye(N);                         % B defines the coordinate system
D = eye(N);                         % diagonal matrix D defines the scaling
C = B*D*(B*D)';                     % covariance matrix
counteval = 0;
%
gen = 0;
global_eval = 0;
arfitness = inf*ones(1,lambda); %Offspring Fitness
bestF = (feval (strfitnessfct,Xmean(:,:))); %Parents' Fitness

aSZ = 0;
iterSZ = 1;

% -------------------- Generation Loop --------------------------------
while global_eval < MAX_EVAL
   
    Xnew = zeros(N,lambda); %temporary DB for offspring.
    mutX = randn(N,lambda); % array of normally distributed r.v.
    for k=1:lambda 
        Xnew(:,k) = Xmean(:,1) + sigma*(B*D*mutX(:,k)); % add mutation
    end
       
    %%%Selection%
    svmupdate = 0;
    if ((global_eval > 100) && (rem(global_eval,1) == 0) && (useSVM == 1))
       svmupdate = 1;

       mutX = randn(N,nTest);% - dX;  % array of normally distributed r.v.
       Xnew = zeros(N,nTest);
       BD = B*D;
       for i=1:nTest,
            Xnew(:,i) = Xmean(:,1) + sigma*(BD*mutX(:,i));
       end;
       xtest = Xnew';
       
       [xtrain, ftrain, nTrain] = SelectTrainPoints(ARX, ARF, aSZ, N, nTrainMax);
  
       ftrain_svm = rand(nTrain,1); 
       ftest_svm = rand(nTest,1);      
       [ftrain_svm, ftest_svm] = GetSVMRank_FAST(xtrain, xtest, nTrain, nTest, N, C, invsqrtC, Xmean, 1, niter);
    
       [ftest_svm, arindexSVM] = sort(ftest_svm,'descend');  % minimization
         
       [selected0, selected] = SelectPointsAccordingToRank(Xnew, N, arindexSVM, nTest, lambda , nfcalc, sigma1, sigma2);
       arxnew = zeros(N,lambda);
       arxnew(:,:) = Xnew(:,arindexSVM(selected0));
       
       Xnew = arxnew;
                  
    end;  
    iterSZ = iterSZ + 1;
    %%%Selection%
    
    nrealcalc = nfcalc;
    if (svmupdate == 0) nrealcalc = lambda;     end;

    counteval = counteval + nrealcalc;
    global_eval = global_eval + nrealcalc;
       
    %update
    if (1)
        ireal = 1;
        fmax = 7e+4;
        for k=1:lambda
            x = Xnew(:,k);

            if (svmupdate == 0)     
                arfitness(k) = feval (strfitnessfct,x);
                ARX(:,aSZ+k) = x;  
                ARF(aSZ+k) = arfitness(k);
            end;
            if (svmupdate == 1)
                    if ( min( abs(selected-k) ) == 0 )
                        f = feval (strfitnessfct,x);
                        arfitness(k) = f;                        
                        ARX(:,aSZ+ireal) = x;       
                        ARF(aSZ+ireal) = f;
                        ireal = ireal + 1;
                    else arfitness(k) = fmax; end; %arfitness(k-1) + k * 1e-5 * abs(arfitness(k-1));    end;
            end; 
        end
                
        if (svmupdate == 1)
            for k=2:nfcalc
                dpos = selected(k)-selected(k-1)-1;
                for k2=1:dpos
                   f0 = arfitness( selected(k-1) );
                   f1 = arfitness( selected(k) );
                   ipos = selected(k-1) + k2;
                   arfitness(ipos) = f0 + (k2*(f1 - f0) / (dpos+1)); 
                end;
            end;
            if (nfcalc > 1)
                    for k=selected(nfcalc):lambda
                        arfitness(k) = arfitness(k-1) + k * 1e-5 * abs(arfitness(k-1));
                    end;
            end;    
        end;
   end;
   %update
    
    aSZ = aSZ + nrealcalc;             
    
    [arfitness2, arindex2] = sort(ARF,'descend');  % minimization
    maxsz = 1100;
    newsz = 1000;
    if (useSVM == 0)
        maxsz = 30;
        newsz = 10;
    end;
    if (aSZ > maxsz)           
        ARX1 = ARX(:,arindex2(aSZ+1-newsz:aSZ));
        ARF1 = ARF(:,arindex2(aSZ+1-newsz:aSZ));
        ARX = ARX1;
        ARF = ARF1;
        aSZ = newsz;
    end;
         
    % Fitness sorting
    [arfitness, arindex] = sort(arfitness,2,'ascend'); %  M I N I M I Z A T I O N
    Xold = Xmean;
    Xmean = Xnew(:,arindex(1:mu)) * weights;  % recombination, new mean value
          
%%%CMA update                    
     % Cumulation: Update evolution paths
      invsqrtC = B * diag( diag(D).^-1) * B';    % C^-1/2
      ps = (1-cs) * ps ... 
              + sqrt(cs*(2-cs)*mueff) * invsqrtC * (Xmean-Xold) / sigma; 
      hsig = norm(ps)/sqrt(1-(1-cs)^(2*counteval/lambda))/chiN < 1.4 + 2/(N+1);
      pc = (1-cc) * pc ...
              + hsig * sqrt(cc*(2-cc)*mueff) * (Xmean-Xold) / sigma; 
          
      % Adapt step size sigma
      sigma = sigma * exp((cs/damps)*(norm(ps)/chiN - 1));

      % Adapt covariance matrix C
      artmp = (1/sigma) * (Xnew(:,arindex(1:mu)) - repmat(Xold,1,mu));
      C = (1-c1-cmu) * C ...                   % regard old matrix  
             + c1 * (pc * pc' ...                % plus rank one update
                     + (1-hsig) * cc*(2-cc) * C) ... % minor correction if hsig==0
             + cmu * artmp * diag(weights) * artmp'; % plus rank mu update 

      % Update B and D from C
      [B,C,D] = update_BCD(B,C,D);      
%%%CMA update
    
    bestF = min(ARF);
    if (bestF < stopfitness)
        break;
    end;

    gen = gen + 1; 
    if (rem(gen,printdelay) == 0)
        disp([ num2str(global_eval) ' ' num2str(bestF) ]);
    end;
end

disp([ num2str(global_eval) ' ' num2str(bestF) ]);
X = Xmean(:,1);
F = bestF;
nevals = counteval;

end
%--------------------------------------------------------------------------
function[new_B,new_C,new_D] = update_BCD(B,C,D);    % Update B and D from C
N = size(D,1);
new_C=triu(C)+triu(C,1)';               % enforce symmetry
[new_B,new_D] = eig(new_C);             % eigen decomposition, B==normalized eigenvectors
new_D(new_D<1E-10)=1E-10; 
if max(diag(new_D)) > 1E14*min(diag(new_D))
    tmp = max(diag(new_D))/1E14 - min(diag(new_D));
	new_C = new_C + tmp*eye(N); new_D = new_D + tmp*eye(N); 
end
new_D = abs(diag(sqrt(diag(new_D))));   % D contains standard deviations now
end

%-------------------------------ACM procedures-------------------------%

%-------------SELECT TRAINING POINTS----------%
function [xtrain, ftrain, nTrain] = SelectTrainPoints(arrX, arrF, aSZ, xdim, nTrainMax)

nTrain = aSZ;
if (nTrain > nTrainMax)   nTrain = nTrainMax;   end;

indx1 = aSZ-nTrain+1:aSZ;
xtrain = arrX(:, indx1)';
ftrain = arrF(:, indx1)';
        
[ftrain, arindex] = sort(ftrain,'ascend');
xtrainnew = zeros(nTrain,xdim);
xtrainnew(:) = xtrain(arindex(1:nTrain),:);
xtrain = xtrainnew;
end

%--------------SVM RANK------------%
function [ftrain_Svm, ftest_Svm] = GetSVMRank_FAST(x_training, x_test, ntraining, ntest, nx, C, invsqrtC, xmean, doEncoding ,niter )

epsilon = 1;
nAlpha = ntraining - 1;
Cvalue = 1e+6;
Ci = zeros(nAlpha,1);
z = 1:nAlpha;
Ci(z) = Cvalue*((nAlpha-z).^2.0 );
sigmaA = 1.0;          
sigmaPow = 1.0;
if (doEncoding == 0)    % 0 - without Encoding          
    invC = C^(-1);
    kernel = 1;         %1 - Mahalanobis RBF(dist = (x1-x2)*invC*(x1-x2)')
end;
if (doEncoding == 1)    % 1 - with Encoding: x'(i) = C^(-0.5) * (x(i)-xmean(i))
    invC = invsqrtC;
    kernel = 0;         %0 - Euclidian RBF
end;
verbose = 0;            % 0 - without comments

[ftrain_Svm0, ftest_Svm0] = RankSVM(x_training', x_test', nx, ntraining, ntest, ...
       niter, epsilon, Ci, kernel, invC, sigmaA, sigmaPow, xmean, doEncoding, verbose);
              
[ftest_Svm0, arindexSVM] = sort(ftest_Svm0,'ascend');
for j=1:ntest
    ftest_Svm0(arindexSVM(j)) = j;
end;

ftrain_Svm = ftrain_Svm0;
ftest_Svm = ftest_Svm0;
end

%-------SELECT POINTS ACCORDING TO RANK VALUE---------%
function [selected1, selected2] = SelectPointsAccordingToRank(arx, xdim, arindexSVM, nPoints, nSelected, nfcalc, sigma1, sigma2)

selected1 = SelecteDistr(floor(nPoints),nSelected, 0, sigma1);
selected2 = SelecteDistr(nSelected, nfcalc, 0, sigma2);
end

%-------SELECT FROM DISTRIBTION--------%
function selected = SelecteDistr(PoolSize, nSelected, mu, sigma)
selected = zeros(nSelected,1);
selected(1) = 1;
for i=2:nSelected
    new = 0;
    while (new == 0)
        new = 1;
        xval = normrnd(mu,sigma);
        newpt = floor( PoolSize*abs(  xval ) );
        if ((newpt < 1) || (newpt > PoolSize)) new = 0;    end;
        if (i > 1)
            [trash, array_position] = min(abs(selected - newpt)); %already exist?
            if (trash == 0) new = 0; end;   
        end;
    end;
    selected(i) = newpt;
end;
[selected, arindex] = sort(selected,'ascend');  % minimization
end
%-----------------------------ACM procedures---------------------------%

%---------------------------BENCHMARK PROBLEMS-------------------------%
function f=frosenbrock(x)
%  if size(x,1) < 2 error('dimension must be greater one'); end
  f = 100*sum((x(1:end-1).^2 - x(2:end)).^2) + sum((x(1:end-1)-1).^2);
end

function f=fsphere(x)
  f=sum(x.^2);
end

function f=fspherenoise(x)
  f = sum(x.^2);
  N = size(x,1);
  e = 0.35 / sqrt(N/2);
  f = f*exp(e*randn());
end

function f=fschwefel(x)
  f = 0;
  sz = size(x,1);
  for i = 1:sz,
    f = f+sum(x(1:i))^2;
  end
end

function f=fschwefel14(x)
  f = fschwefel(x) ^ 0.25;
end

function f=fackley(x)
n = size(x,1);
a = 20; b = 0.2; c = 2*pi;
s1 = 0; s2 = 0;
for i=1:n;
   s1 = s1+x(i)^2;
   s2 = s2+cos(c*x(i));
end
f = a - a*exp(-b*sqrt( (1/n)*s1 )) - exp( (1/n)*s2 ) + exp(1);
end

function f=felli(x)
  N = size(x,1); if N < 2 error('dimension must be greater one'); end
  f=1e6.^((0:N-1)/(N-1)) * x.^2;
end

function f=frastrigin(x)
%  N = size(x,1); if N < 2 error('dimension must be greater one'); end
  scale=1;%10.^((0:N-1)'/(N-1));
  sz = size(x,1);
  f = 10*sz + sum((scale.*x).^2 - 10*cos(2*pi*(scale.*x)));
end